/*
 * Copyright (c) 1983 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Berkeley.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef lint
static char sccsid[] = "@(#)acu.c	5.6 (Berkeley) 9/2/88";
#endif /* not lint */

#include "tip.h"

static sigfunc_t acuabort();

boolean conflag = NO; /* connect flag */

extern int	slip;			/* bring up SLIP on the line */
static jmp_buf jmpbuf;


/*
 * Establish connection for tip
 *
 * If DU is true, we should dial an ACU whose type is AT.
 * The phone numbers are in PN, and the call unit is in CU.
 *
 * If the PN is an '@', then we consult the PHONES file for
 *   the phone numbers.  This file is /etc/phones, unless overriden
 *   by an exported shell variable.
 *
 * The data base files must be in the format:
 *	host-name[ \t]*phone-number
 *   with the possibility of multiple phone numbers
 *   for a single host acting as a rotary (in the order
 *   found in the file).
 */
/* tip+
 * CU is discarded.
 *
 * Phones format sample:
 * HostName PN1,PN2,...,PNn
 * PN is modem dependent.
 *    0=052-809-1234
 *    = pause a second
 * -Kenar-
 */
 
/*
 * lconnect:
 *   take logs
 *   enter tip+ script
 *      dial up is done in tip+ script
 * return:
 *   NULL if connect
 *   error message, if fail
 * -Kenar-
 */
char *lconnect()
{
	char *msg,acupath[PATHSIZ];
	struct stat statbuf;
	char *PN_; /* last call to record when failed. */
	char *host = svalue(HOST);
	sigfunc_t (*f_int)();
	
	loginit(svalue(LOG)); /* moved from tip.c -Kenar-*/

	if (!DU) { /* regular connect message */
		if (CM != NOSTR)
			write(FD, CM, size(CM));
		logent(host, "", DV, "call completed");
		conflag = YES;
		return (NOSTR);
	}

	/* tip+ */
	sprintf(acupath,ACUFMT,AT);
	if(stat(acupath,&statbuf) != 0) return ("unknown ACU type");

	f_int = signal(SIGINT, acuabort);
	if (setjmp(jmpbuf)) {
		signal(SIGINT, f_int);

		if (conflag) msg = NOSTR;
		else msg = "call aborted";
		disconnect(host, msg);
		return ("call aborted");
	}
	
	
   /*
	* Example: PN="052-809-1234,052-809-4321"
	* Note that PN=@ case is preprocessed in getremcap().
	* -Kenar-
	*/
	for(PN = getoken(PN,","); PN; PN = getoken(NULL,",")){
		/* tip+ */
		iflush(FD);
		msg = login(acupath);
		if(!msg){ /* success */
			signal(SIGINT, f_int);
			PN = makstr(PN);
			logent(host, PN, AT,"call completed");
			conflag = YES;
			return (NOSTR);
		}
		
		/* reset modem to command mode */
		clrdtr(FD); /* clear data terminal ready */
		nap(500);
		setdtr(FD); /* set data terminal ready */
		nap(500);
		PN_ = PN;
	}
	signal(SIGINT, f_int);
	PN = PN_; /* come back to last call for log. */
	if(equal(msg,"")) msg = "no dialer";
	else if(equal(msg,"-")) msg = "call failed";
	disconnect(host, msg);
	return (msg);
}


void disconnect(host,reason)
	char *host,*reason;
{
	if(DU){
		if (reason) logent(host, PN, AT, reason);
		else logent(host, PN, AT, "call terminated");
	}
	else{
		if (reason) logent(host, "", DV, reason);
		else logent(host, "", DV, "call terminated");
	}
	clrdtr(FD);
}

static sigfunc_t
acuabort()
{
	longjmp(jmpbuf, 1);
}
