/*
 * Copyright (c) 1983 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Berkeley.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef lint
static char sccsid[] = "@(#)cmds.c	5.8 (Berkeley) 9/2/88";
#endif /* not lint */

#include "tip.h"

#define EOFTIME 1  /* EOFTIME idle means EOF from remote */

/* Arguments for cmd_child
 * We use such like: cmd_child(FREEZE)
 */
#define FREEZE   1
#define UNFREEZE 2
#define SWEEP    3
#define UNSWEEP  4
#define WAITSYNC   5

/*
 * tip
 *
 * miscellaneous commands
 */

void variable();
void vcheck();

static void xsend(char c);
static sigfunc_t intprompt(); /* used in handling SIG_INT during prompt */
static sigfunc_t sigint();	/* interrupt routine for file transfers */
static void transfer(char *cmd, int fd);
static void transmit(FILE *fd, char *eofchars, char *command);
static void usrftp(char *remote,char *local);
static void execute(char *comline);

extern struct sgttyb	arg;		/* current mode of local terminal */
extern struct tchars	tchars;		/* current state of terminal */
extern struct tchars	defchars;	/* initial state of terminal */
extern int	fildes[2];		/* file transfer synchronization channel */
extern int	repdes[2];		/* read process sychronization channel */
extern char	ccc;			/* synchronization character */
extern int	pid;			/* pid of tipout */
extern int suidmode; /* defined in tip.c */
static int	quant[] = { 60, 60, 24 };
static char *sep[] = { "second", "minute", "hour" };
static void sigIOT(char *s);
static boolean xquit;



/*
 * Help command
 */
void help(c) char c;
{
	register esctable_t *p;
	extern esctable_t etable[];

	for(p = etable; p->e_char; p++){
		printf("%2s",ctrl(cvalue(ESCAPE)));
		printf("%-2s   %s\r\n",ctrl(p->e_char),p->e_help);
	}
}


static jmp_buf promptbuf;

/*
 * Interrupt service routine during prompting
 */
static sigfunc_t intprompt()
{
	longjmp(promptbuf,1);
}


/*
 * Print string ``s'', then read a string
 *  in from the terminal.  Handles signals & allows use of
 *  normal erase and kill characters.
 */
static boolean prompt(s,p)
char *s;
char *p;
{
	char *b = p;
	sigfunc_t(*oint)();
	boolean stopped;

	stopped = 0;
	oint = signal(SIGINT,intprompt);
	tio_unraw();
	printf("%s",s);
	if(setjmp(promptbuf)){
		stopped = 1;
		printf("\r\n");
	}
	else{
		while((*p = getchar()) != EOF && *p != '\n')
			p++;
	}
	
	*p = '\0';
	tio_raw();
	signal(SIGINT,oint);
	return(stopped || p == b);
}

/*
 * Escape handler --
 *  called on recognition of ``escapec'' at the beginning of a line
 */
int escape()
{
	register char gch;
	register esctable_t *p;
	char c = cvalue(ESCAPE);
	extern esctable_t etable[];

	gch = (getchar()&0177);
	for(p = etable; p->e_char; p++)
		if(p->e_char == gch){
			(*p->e_func)(gch);
			return(0);
		}
	/* ESCAPE ESCAPE forces ESCAPE */
	if(c != gch)
		pwrite(&c,1);
	return(gch);
}

/* tip+3.0
 * sigint: SIGINT function to quit file operations
 * -Kenar-
 */

static sigfunc_t sigint()
{	puts("INTERRUPT");
	xquit = YES;
}


/* tip+3.0
 * cmd_child: child control commands
 * -Kenar-
 */
static void cmd_child(cmd)
int cmd;
{	
	switch(cmd){
	case FREEZE:
		sigIOT(CWAIT);
		read(repdes[0], (char *)&ccc, 1); /* Wait until read process stops */
		break;
	case UNFREEZE:
		write(fildes[1], (char *)&ccc, 1);
		break;
	case SWEEP:
		sigIOT(CSWEEPON);
		break;
	case UNSWEEP:
		sigIOT(CSWEEPOFF);
		break;
	case WAITSYNC:
		if(kill(pid,SIGUSR1)) printf("SIGUSR1 not caught");
		break;
	default:;
	}
}

/* tip+3.0
 * subfname: substitute '?' by file name
 * EXAMPLE
 * subfname(buf,"cat >?",somefile) makes buf="cat >somefile"
 * coded by Kenar
 */
static void subfname(buf,fmt,fname)
char *buf,*fmt,*fname;
{	char *p,*q;
	q = buf;
	p = fmt;
	while(*p && *p != '?') *q++ = *p++;
	*q = '\0';
	if(*p == '?'){
		q = strapp(q,fname);
		q = strapp(q,++p);
	}
}
	


/*
 * FTP - remote ==> local
 *  get a file from the remote host
 */
void getfl(c) char c;
{
	char buf[256], *cp;
	char fname[PATHSIZ];
	int	sfd;
	
	putchar(c);
	/*
	 * get the UNIX receiving file's name
	 */
	if (prompt("Local file name to receive ? ", fname))
		return;
	cp = expand(fname);
	if ((sfd = creat(cp, 0666)) < 0) {
		printf("\r\n%s: cannot creat\r\n", cp);
		return;
	}
	
	/*
	 * collect parameters
	 */
	if (prompt("List command for remote system ? ", buf)) {
		unlink(cp);
		return;
	}
	transfer(buf, sfd);
	close(sfd);
}


/*
 * Cu-like take command
 */
/*ARGUSED*/
void cu_take(cc) char cc;
{
	int fd, argc;
	char *fnamex;
	char fname[PATHSIZ];
	char *argv[10];
	char buf1[STRSIZ];

	if(vndef(RTAKE) && vndef(LTAKE)){
		puts("Rtake or ltake is undefined.\r");
		return;
	}
	if(index(svalue(RTAKE),'?') || index(svalue(LTAKE),'?')){
		if (prompt("[take] ", fname)) return;
		if ((argc = args(fname, argv)) < 1 || argc > 2) {
			printf("usage: <take> from [to]\r\n");
			return;
		}
		if (argc == 1) argv[1] = argv[0];
	}
	else argv[0] = argv[1] = "";
	
	subfname(buf1,svalue(RTAKE),argv[0]);
		/* example: svalue(RTAKE)=="cat ?;sleep 2" */
	fnamex = expand(argv[1]);
	if(vdef(LTAKE)){
		char buf2[STRSIZ];
		subfname(buf2,svalue(LTAKE),fnamex);
		usrftp(buf1,buf2);
		return;
	}
	
	if ((fd = creat(fnamex, 0666)) < 0) {
		printf("\r\n%s: cannot create\r\n", fnamex);
		return;
	}
	transfer(buf1, fd);
	close(fd);
}

/*
 * FTP - remote ==> local process
 *   send remote input to local process via pipe
 */
pipefile()
{
	int cpid, pdes[2];
	char buf[256];
	int status, p;

	if (prompt("Local command? ", buf))
		return;

	if (pipe(pdes)) {
		printf("can't establish pipe\r\n");
		return;
	}

	if ((cpid = fork()) < 0) {
		printf("can't fork!\r\n");
		return;
	} else if (cpid) {
		if (prompt("List command for remote system? ", buf)) {
			close(pdes[0]), close(pdes[1]);
			kill (cpid, SIGKILL);
		} else {
			close(pdes[0]);
			transfer(buf, pdes[1]);
			close(pdes[1]);
			while ((p = wait(&status)) > 0 && p != cpid)
				;
		}
	} else {
		register int f;

		dup2(pdes[0], 0);
		close(pdes[0]);
		for (f = 3; f < 20; f++)
			close(f);
		execute(buf);
		printf("can't execl!\r\n");
		exit(0);
	}
}

#ifdef NODEF
/* ---------------- WARNING ------------
 * You may be prompted to use 'popen()' for simplicity.
 * However 'popen()' makes security hole in pseudo-user execution.
 * We keep this codes here for precaution
 * and for the day tip+ establish rigorous user execution .
 * -Kenar-
 */

/*
 * FTP - remote ==> local process
 *   send remote input to local process via pipe
 *
 * tip+ (Kenar)
 * 1996/01 coded by Kenar
 */
void pipefile()
{
	char buf1[STRSIZ],buf2[STRSIZ];
	int fd;
	FILE *fp;

	if (prompt("Local command? ", buf1)) return;
	if (prompt("List command for remote system ? ", buf2)) return;

	fp = popen(buf1,"w");
	if(!fp){
		puts("Local command failed.");
		return;
	}
	fd = fileno(fp);

	/*
	 * Be careful not to send garbage to fd. Impossible to truncate!
	 * I have a good idea.
	 * We execute: (command; sleep 2)
	 * -Kenar-
	 */
	sprintf(buf1,"%s",buf2);
	transfer(buf1, fd);
	pclose(fp);
}
#endif

static jmp_buf cjmpbuf; 

static sigfunc_t intcbreak()
{
	longjmp(cjmpbuf,1);
}


/* tip+3.0
 * conexe: connect remote tty to local tty and execute command
 * coded by Kenar
 */
static void conexe(comln)
char *comln;
{	int cpid, status, p;
	time_t start;
	sigfunc_t (*int_old)();

	tandem(NO);
	
	/*
	 * Set up file descriptors in the child and
	 *  let it go...
	 */
	cpid = fork();
	if (!cpid) { /* child process */
		int i;
		dup2(FD, 0);
		dup2(FD, 1);
		for (i = 3; i < 20; i++) close(i);
		signal(SIGINT, SIG_DFL);
		execute(comln);
		printf("can't execute `%s'\r\n", comln);
		exit(0);
	}
	start = time(0);
	tio_cbreak(YES);
	int_old = signal(SIGINT,intcbreak);
	if(setjmp(cjmpbuf)) kill(cpid,SIGTERM);
	while ((p = wait(&status)) > 0 && p != cpid);
	prtime("away for ", time(0)-start);
	tandem(bvalue(TAND));
	tio_cbreak(NO);
	signal(SIGINT, int_old);
}
	


/* tip+3.0
 * usrftp: user defined ftp
 * coded by Kenar
 */
static void usrftp(remote,local)
char *remote; /* remote command */
char *local;  /* local command */
{	char c;

	pwrite(remote, size(remote));
	sleep(1);
	cmd_child(FREEZE);
	/*
	 * finish command
	 */
    /* processing echoes of command
	 * Have a knowledge. The sequence is: \r\n
	 * -Kenar-
	 */
	nap(100);
	iflush(FD); /* flush FD buffer to remove echo garbage */
	if(*remote){
		pwrite("\r", 1);
		repeat {
			pread(&c, 1);
		} until (c == '\n');
	}
	
	conexe(local);

	cmd_child(UNFREEZE);	
}
	


/*
 *  Bulk transfer routine --
 *  used by getfl(), cu_take(), and pipefile()
 *
 * tip+ (kenar)
 * 1996/01 Replaced by new algorithm and changed the usage.
 * Now EOF in transfer is EOFTIME second idle time.
 * -Kenar-
 */
static void transfer(cmd, fd)
	char *cmd; /* command for remote host */
	int fd;    /* file descripter to be written */
{
	char c, buf[BUFSIZ], *frbuf;
	register char *p,*q;
	int n,ccount,lcount,frsiz,len,m;
	time_t start;
	sigfunc_t (*f)();
	char xeof = cvalue(EOFREAD);
	char *xctl = svalue(XCONTROL);
	char xabort = xctl[1];
	char xskip = xctl[3];
	
	xabort = (xabort == '-' ?  0 : CTRL(xabort));
	xskip = (xskip == '-' ?  0 : CTRL(xskip));
	
	
	frsiz = nvalue(FRAMESIZE);
	frbuf = malloc(frsiz * sizeof(char));
	if(!frbuf){
		puts("Can't allocate frame buffer.\r");
		return;
	}

	pwrite(cmd, size(cmd));

	cmd_child(FREEZE);

	/*
	 * finish command
	 */
    /* processing echoes of command
	 * Have a knowledge. The sequence is: \r\n
	 * -Kenar-
	 */
	nap(100);
	iflush(FD); /* flush FD buf to remove echo garbage */
	pwrite("\r", 1);
	repeat {
		pread(&c, 1);
	} until (c == '\n');
	
	/* enable ^C */
	tio_cbreak(YES);
	f = signal(SIGINT, sigint);

	start = time(0);
	
	m = 0;
	xquit = NO;
	lcount = ccount = 0;
	p = frbuf;
	
	pread(p,1); /* wait until host makes effect */
	if(*p == '\r') lcount++;
	else{
		ccount++;
		write(fd,p,1);
	}
	
	for (;!xquit;){
		len = nread(FD);
		if(len == 0){ /* If EOFTIME idle time then exit */
			nap(50);
			m++;
			if( m == 20*EOFTIME ) break; 
			continue;
		}
		m = 0;
		
		n = pread(buf,BUFSIZ);
		if( n <= 0 ) break;

		ccount += n;
		q = buf;
		while(n--){
			if(xabort && *q == xabort) break;
			if(xeof  && *q == xeof) break;
			if(*q == '\r'){	 /* skip '\r' */
				lcount++; q++;
				fprintf(stderr," %d\r",lcount);
			}
			else *p++ = *q++;
			if(p == frbuf + frsiz){
				if(bvalue(TAND)) pwrite("\023",1); /* send ^S */
				write(fd,frbuf,frsiz);
				if(bvalue(TAND)) pwrite("\021",1); /* send ^Q */
				p = frbuf;
			}
		}
	}
	
	write(fd,frbuf, p - frbuf);

	printf(" %d lines and %d chars are transferred in %d seconds.\r",
		lcount,
		ccount,
		time(0)-start);
		
	/* disable ^C */
	tio_cbreak(NO);
	signal(SIGINT, f);
	
	cmd_child(UNFREEZE);
	free(frbuf);
	if(xskip) pwrite(&xskip, 1);
}


/*
 * Cu-like put command
 */
/*ARGUSED*/
cu_put(cc)
	char cc;
{
	FILE *fd;
	char rline[STRSIZ];
	int argc;
	char *argv[10];
	char *fnamex;
	char fname[PATHSIZ];

	if(vndef(RPUT) && vndef(LPUT)){
		puts("Both rput and lput are undefined\r");
		return;
	}
	if(index(svalue(RPUT),'?') || index(svalue(LPUT),'?')){
		if (prompt("[put] ", fname)) return;
		if ((argc = args(fname, argv)) < 1 || argc > 2) {
			printf("usage: <put> from [to]\r\n");
			return;
		}
		if (argc == 1) argv[1] = argv[0];
	}
	else argv[0] = argv[1] = "";
	
	subfname(rline, svalue(RPUT), argv[1]);
	fnamex = expand(argv[0]);
	if(vdef(LPUT)){
		char lline[STRSIZ];
		subfname(lline, svalue(LPUT), fnamex);
		usrftp(rline,lline);
		return;
	}
	
	if ((fd = fopen(fnamex, "r")) == NULL) {
		printf("%s: cannot open\r\n", fnamex);
		return;
	}
	
	transmit(fd,svalue(EOFWRITE), rline);
	fclose(fd);
}

/*
 * FTP - local ==> remote
 *  send local file to remote host ( non-unix )
 *  terminate transmission with pseudo EOF sequence
 */
/* tip+
 * 1993/08 changed message
 */
sendfile(cc)
	char cc;
{
	FILE *fd;
	char *fnamex;
	char fname[PATHSIZ];

	putchar(cc);
	/*
	 * get file name
	 */
	if (prompt("Local file name to send ? ", fname))
		return;

	fnamex = expand(fname);
	if ((fd = fopen(fnamex, "r")) == NULL) {
		printf("%s: cannot open\r\n", fnamex);
		return;
	}
	transmit(fd, svalue(EOFWRITE), NULL);
	fclose(fd);
}


/* ---------------- WARNING ------------
 * You may be prompted to use 'popen()' for simplicity.
 * However 'popen()' makes security hole in pseudo-user execution.
 * We keep this codes here for precaution
 * and for the day tip+ establish rigorous user execution .
 *
 * pipeout: pipe out local command to remote host
 * 1996/01 Replaced by Kenar
 * -Kenar-
 */
void pipeout(c)
char c;
{
	char buf[256],*cmd;
	FILE *fd;

	if(suidmode){
		puts("Currentlly unsuppopted.\r"); /* for security reason. */
		return;
	}
	
	putchar(c);
	if (prompt("Local command? ", buf)) return;
		
	cmd = expand(buf);
	if ((fd = popen(cmd, "r")) == NULL) {
		printf("%s: cannot execute\r\n", cmd);
		return;
	}
	transmit(fd, svalue(EOFWRITE), NULL);
	pclose(fd);
}		


#ifdef NODEF
/*
 * Stolen from consh() -- puts a remote file on the output of a local command.
 *	Identical to consh() except for where stdout goes.
 */
/*
 * Temporally I use this with a little modification.
 * But somethig wrong!
 * -Kenar-
 */
pipeout(c)
{
	char buf[256];
	int cpid, status, p;
	time_t start;
	sigfunc_t (*int_old)();

	putchar(c);
	if (prompt("Local command? ", buf))
		return;
		
	cmd_child(SWEEP);
	
	int_old = signal(SIGINT, SIG_IGN);
	tio_cbreak(YES);
	
	/* cmd_child(FREEZE); */
	
	/*
	 * Set up file descriptors in the child and
	 *  let it go...
	 */
	if ((cpid = fork()) < 0)
		printf("can't fork!\r\n");
	else if (cpid) {
		start = time(0);
		while ((p = wait(&status)) > 0 && p != cpid)
			;
	} else {
		register int i;

		dup2(FD, 1);
		for (i = 3; i < 20; i++)
			close(i);
		signal(SIGINT, SIG_DFL);
		execute(buf);
		printf("can't find `%s'\r\n", buf);
		exit(0);
	}

	prtime("away for ", time(0)-start);
		
	cmd_child(UNSWEEP);
	/* cmd_child(UNFREEZE); */
	tio_cbreak(NO);
	signal(SIGINT, int_old);
}
#endif


static char synch; /* synch is used in tramsmit and xsend. */


/*
 *   Bulk transfer routine to remote host --
 *   used by sendfile() and cu_put()
 *
 *   1996/01 Reconstructed (tip+3.0)
 *   Changed method of transmission to remove buggy actions.
 *   Now echoes are received and sweep out by child.
 *   Synchronization if necessary.
 *
 *   NO SYNC MODE PROTOCOL
 *   messages to child:
 *      transmit mode on          --- cmd_child(SWEEP)
 *      text transmission
 *      transmit mode off         --- cmd_child(UNSWEEP)
 *
 *   SYNC MODE PROTOCOL
 *   messages to child:
 *      transmit mode on          --- cmd_child(SWEEP)
 *      line transmission ( no '\r' )
 *      wait to be send synch   --- cmd_child(WAITSYNC) = kill(pid,SIGUSR1)
 *      send '\r'
 *      sleep waiting SIGALRM from child
 *   messages from chile:
 *      SIGALRM ( got synch )     --- kill(getppid(),SIGALRM) in tipout.c
 *   -Kenar-
 */
static void transmit(fd, eofchars, command)
	FILE *fd;
	char *eofchars, *command;
{
	char *pc,lastc;
	int c, ccount, lcount, col;
	int syncmode;
	int omask,umask,vmask;
	time_t start_t;
	sigfunc_t (*int_old)();
	char receipt = svalue(XCONTROL)[2];
	
	receipt = ( receipt == '-' ? 0 : CTRL(receipt));
	syncmode = bvalue(LINESYNC) ? YES : NO;
	cmd_child(SWEEP);

	/* enable ^C */
	int_old = signal(SIGINT, sigint);
	tio_cbreak(YES);

	lastc = '\0';
	start_t = time(0);
	
	synch = '\r';

   /*
	* I use SIGUSR2 only for '\r' sync
	*/
	umask = sigmask(SIGALRM);
	omask = sigblock(umask);
	vmask = omask & ~umask;
	
	
	ccount = 0;
	lcount = 0;
	col = 0; /* column to expand tab */
	
	if (command != NULL) {
		for (pc = command; *pc; pc++) xsend(*pc);
		pwrite("\r",1);
		sleep(3); /* wait for remote stty to take effect */
	}

	puts("\r");
	xquit = NO;
	for(;!xquit;) { /* xquit: turn to YES by SIGINT */
		c = getc(fd);
		if (c == EOF) break;
		lastc = c;
		if(c < 040){
			if (c == '\n') c = '\r';
			else if(!index(svalue(EXCEPTIONS),c)) continue;
			else if (c == '\t' && bvalue(TABEXPAND)) {
				xsend(' ');
				while ((++col % nvalue(TABSIZE)) != 0) xsend(' ');
				continue;
			}
		}
		if(c == synch){ /* at eol */
			ccount += col;
			col = 0;
			lcount++;
			fprintf(stderr," %d\r", lcount);
			if(syncmode){
				nap(10); /* wait so that child is ready */
				cmd_child(WAITSYNC);
				sigblock(vmask);
				ccount++; /* See note bellow */
				xsend(c); /* LDELAY nap in xsend that is after pwrite */
				sigblock(umask);
				continue;
			}
		}
		col++;
		ccount++; /* See note bellow */
		xsend(c); /* LDELAY nap in xsend that is after pwrite */
	   /*
	    * nap(0) clear nap timer
		* nap()  discards SIGALRM
		* -Kenar-
		*/
	}

	if (lastc != '\n') xsend('\r');
	for (pc = eofchars; *pc; pc++) xsend(*pc);
	/* NOTE ^D is effective only after \r */

	sleep(1);
	cmd_child(UNSWEEP);
	printf("\r\n %d lines and %d chars are transferred in %d second.\r\n",
		lcount,
		ccount,
		time(0)-start_t);
	if(receipt){
		char cbuf[1];
		nap(100);
		repeat pread(cbuf,1);
		until(cbuf[0] == receipt);
	}
	/* disable ^C */
	signal(SIGINT, int_old);
	sigblock(omask);
	tio_cbreak(NO);
	pwrite("\r", 1);
	return;
}


/*
 * FTP - send single character
 *  xsend is used only in transmit()
 */
static void xsend(char c)
{
	pwrite(&c, 1);
	if(c == synch){
		if(nvalue(LDELAY) > 0) nap(nvalue(LDELAY));
	}
	else if(nvalue(CDELAY) > 0) nap(nvalue(CDELAY));
}



/*
 * Fork a program with: (original form)
 *  0 <-> local tty in
 *  1 <-> local tty out
 *  2 <-> local tty out
 *  3 <-> remote tty in
 *  4 <-> remote tty out
 *
 * Fork a program with: (tip+3.0 form)
 *  0 <-> remote tty in
 *  1 <-> remote tty out
 *  2 <-> local tty out (stderr)
 */
consh(c)
{
	char buf[256];
	putchar(c);
	if (prompt("Local command? ", buf))
		return;
	cmd_child(FREEZE);
	
	nap(100);
	iflush(FD); /* flush FD buffer to remove echo garbage */

	conexe(buf);

	cmd_child(UNFREEZE);
}


/*
 * filexfer_help
 *      tipx
 *      tip+ mod by Kenar
 *
 * Display help for ~X
 *   type = 0, general help, 1 = send help, 2 = receive help
 */
filexfer_help(type)
int type;
{
static char *ghelp[] =
{
	"for send help, type ~Xs?; for receive help, type ~Xr?",
	"where ~ represents your escape character",
	(char *)0,
};
static char *shelp[] =
{
	"XMODEM (CHK, CRC) and XMODEM-1k/CRC",
	"    sx [-ak] filename",
	"        -a (ASCII) convert NL to CR/LF (default binary)",
	"        -k use 1024 blocks instead of 128 (XMODEM-1k)",
	"YMODEM/CRC Batch",
	"    sy [-af] filelist",
	"        -a (ASCII) convert NL to CR/LF (default binary)",
	"        -f transfer using full pathanme (default simple)",
	"ZMODEM (CRC-16, CRC-32)",
	"    sz [-af+nNyrwo] [-w #] [-L #] [-l #]  filelist",
	"        -a (ASCII) convert NL to CR/LF (default binary)",
	"        -f transfer using full pathanme (default simple)",
	"        -+ Append to existing destination file",
	"        -n overwrite file if source newer",
	"        -N overwrite file if source newer or longer",
	"        -y yes, absolutely overwrite existing file",
	"        -r Resume/Recover interrupted file transfer",
	"        -o use CRC-16 instead of CRC-32",
	"        -w # Window is # bytes (>= 256, multiple of 64)",
	"        -L # Limit subpacket length to # bytes",
	"        -l # Limit frame length to # bytes (l must >= L)",
	"    Not all ZMODEM receivers accept all overwrite or append options."
	"B+",
	"    bp",
	(char *)0
};
static char *rhelp[] =
{
	"XMODEM (CHK, CRC) (rcvr tries CRC, then checksum)",
	"    rx [-ab] filename",
	"YMODEM/CRC Batch",
	"    ry [-abu]",
	"ZMODEM (CRC-16, CRC-32)",
	"    rz [-abu]",
	"Switches:    -a force ASCII translation on receive",
	"             -b force binary transfer",
	"             -u convert uppercase filenames to lower case",
	"B+",
	"    bp",
	(char *)0
};
char **hh;

	switch(type)
	{
		case 0:  hh = ghelp; break;
		case 1:  hh = shelp; break;
		case 2:  hh = rhelp; break;
	}
	while(*hh)
	{
		fputs(*hh++,stdout);
		fputs("\r\n",stdout);
	}
}


/*
 * filexfer
 *      tipx
 *      tip+ mod by Kenar
 *
 * Fork a file transfer with:
 *  0 <-> local tty in
 *  1 <-> local tty out
 *  2 <-> local tty out
 *  FD <-> remote tty in/out
 */
void filexfer(c)
{
	char buf[256],ch;
	char xcmd[256+24];
	int cpid, status, p, err;
	time_t start;
	sigfunc_t (*int_old)();
	
	/* added by Kenar */
	if(equal(svalue(XFER),"bp")){
		sprintf(xcmd,"tipbp -I -. %d",FD);
		strcpy(buf,"bp");
	}
	else{
		puts("File transfer command. Blank line will quit this mode.\r");
		puts("    r: receive\r");
		puts("    s: send\r");
		err = 1;
		do
		{	puts("Select --- \r");
			read(0,&ch,1);
			if(isupper(ch)) ch = tolower(ch);
			if(isspace(ch)) return;
			if(index("rs",ch)) err = 0;
			else	puts("Erorr: Invalid selection.\r");
		} while(err);
	
		if(ch == 'r')
		{	puts("usage is ...\r");
			filexfer_help(2);
			prompt("command: ",buf);
		}
		else
		{	puts("usage is ...\r");
			filexfer_help(1);
			prompt("command: ",buf);
		}
	
		if(!strncmp(buf,"sz ",3))
			sprintf(xcmd,"tipsz -Z -. %d",FD);
		else if(!strncmp(buf,"sy ",3))
			sprintf(xcmd,"tipsz -Y -k -. %d",FD);
		else if(!strncmp(buf,"sx ",3))
			sprintf(xcmd,"tipsz -X -. %d",FD);
		else if(!strncmp(buf,"rz",2))
			sprintf(xcmd,"tiprz -Z -. %d",FD);
		else if(!strncmp(buf,"ry",2))
			sprintf(xcmd,"tiprz -Y -. %d",FD);
		else if(!strncmp(buf,"rx ",3))
			sprintf(xcmd,"tiprz -X -. %d",FD);
		else if(!strncmp(buf,"bp",2))
			sprintf(xcmd,"tipbp -I -. %d",FD);
		else{
			puts("unrecognized transfer command\r");
			return;
		}
	}

	strcat(xcmd,buf + 2);

	int_old = signal(SIGINT, SIG_IGN);
	tio_unraw();
	cmd_child(FREEZE);

	if ((cpid = vfork()) < 0)
		printf("can't fork!\r\n");
	else if (cpid) {
		start = time(0);
		while ((p = wait(&status)) > 0 && p != cpid)
			;
	} else {
		register int i;

		for (i = 3; i < 20; i++)
		{
			if(i != FD)
				close(i);
		}
		signal(SIGINT, SIG_DFL);
		execute(xcmd);
		printf("tip: can't execute `%s'\n", xcmd);
		exit(0);
	}
	prtime("away for ", time(0)-start);
	cmd_child(UNFREEZE);
	tio_raw();
	signal(SIGINT, int_old);
}

/*
 * Escape to local shell
 */
shell()
{
	int shpid, status;
	extern char **environ;
	char *cp;
	sigfunc_t (*int_old)();

	printf("[sh]\r\n");
	int_old = signal(SIGINT, SIG_IGN);
	tio_unraw();
	if (shpid = fork()) {
		while (shpid != wait(&status));
		tio_raw();
		printf("\r\n!\r\n");
		signal(SIGINT, int_old);
		return;
	} else {
		signal(SIGINT, SIG_DFL);
		if ((cp = rindex(svalue(SHELL), '/')) == NULL)
			cp = svalue(SHELL);
		else
			cp++;
		shell_uid(); execle(svalue(SHELL), cp, 0,environ);
		printf("\r\ncan't execle!\r\n");
		exit(1);
	}
}


/*
 *  Send signal to reset variable in TIPOUT
 *  Original function setscript() is replaced by this function.
 *	-Kenar-
 *	The format of piped data is,
 *	vid:value
 *	data is NULL terminated.
 * 	EXAMPLE
 *	1:1		BEAUTIFY:TRUE
 *	2:9600	BAUDRATE:9600
 *  - Kenar -
 */
void sigEMT(vid) int vid; /* vid: message header = variable ID */
{
	int n = 0;
	int status;
	char buf[256];
	/*
	 * enable TIPOUT side for dialogue
	 */

	repeat
	{	status = kill(pid,SIGEMT);
		nap(20);
		n++;
		if(n > 5) puts("signal unreachable!\r");
	} until(status == 0);
	
	switch(vtype(vid)&TMASK){
	case STRING:
		sprintf(buf,"%d:%s",vid,svalue(vid));
		break;
	case NUMBER:
		sprintf(buf,"%d:%d",vid,nvalue(vid));
		break;
	case BOOL:
		sprintf(buf,"%d:%d",vid,bvalue(vid));
		break;
	case CHAR:
		sprintf(buf,"%d:%c",vid,cvalue(vid));
		break;
	}
	write(fildes[1], buf, strlen(buf) + 1);
}

static void sigIOT(s) char *s;
{
	int n = 0;
	int status;
	/*
	 * enable TIPOUT side for dialogue
	 */

	repeat
	{	status = kill(pid,SIGIOT);
		nap(20);
		n++;
		if(n > 5) puts("signal unreachable!\r");
	} until(status == 0);
	
	write(fildes[1], s, strlen(s) + 1);
}




/*
 * Change current working directory of
 *   local portion of tip
 *
 * 1995/12 tip+
 * Kenar says,
 * It is unfriendly for beginner who may push ~c with no knowledge and
 * may be brought to HOME directory.
 * So I changed usage:
 * no field in
 * [cd]
 * makes no effect.
 * -Kenar-
 */
void chdirectory()
{
	char dirname[80];
	char *cp;
	if (prompt("[cd] ", dirname)) return;
	cp = expand(dirname);
	if (chdir(cp) < 0)
		printf("%s: bad directory\r\n", cp);
	printf("!\r\n");
}

finish()
{
	char *dismsg;
	
	if(vdef(DISCONNECT)){
		dismsg = svalue(DISCONNECT);
		pwrite(dismsg, strlen(dismsg));
		sleep(5);
	}
	kill(0,SIGTERM);
}

static void execute(comline) char *comline;
{
	shell_uid(); execl("/bin/sh", "sh", "-c", comline, 0);
}

args(buf, a)
	char *buf, *a[];
{
	register char *p = buf, *start;
	register char **parg = a;
	register int n = 0;

	do {
		while (*p && (*p == ' ' || *p == '\t'))
			p++;
		start = p;
		if (*p)
			*parg = p;
		while (*p && (*p != ' ' && *p != '\t'))
			p++;
		if (p != start)
			parg++, n++;
		if (*p)
			*p++ = '\0';
	} while (*p);

	return(n);
}

prtime(s, a)
	char *s;
	time_t a;
{
	register i;
	int nums[3];

	for (i = 0; i < 3; i++) {
		nums[i] = (int)(a % quant[i]);
		a /= quant[i];
	}
	printf("%s", s);
	while (--i >= 0)
		if (nums[i] || (i == 0 && nums[1] == 0 && nums[2] == 0))
			printf("%d %s%c ", nums[i], sep[i],
				nums[i] == 1 ? '\0' : 's');
	printf("\r\n!\r\n");
}


/*
 * The action of this function is a pair of intEMT in tipout.c
 * -Kenar-
 */
void variable()
{
	char buf[256];
	
	if (prompt("[set] ", buf))
		return;
	vlex(buf);
	vcheck();
}


/*
 * Turn tandem mode on or off for remote tty.
 */
tandem(cntl)
boolean cntl;
{
	io_tandem(FD,cntl);
}

/*
 * Send a break.
 */
genbrk()
{	sendbrk(FD,nvalue(BRKSIZ));
}


/*
 * control rawthru mode of forked process
 *      tipx
 *      tip+ mod by Kenar
 */
rawthru_control(gch)
char gch;
{	extern int rawthru;
	gch &= 0177;
	if(gch == 'R')
	{
		rawthru = 1;
		sigIOT(CRTHON);
	}
	else
	{
		rawthru = 0;
		sigIOT(CRTHOFF);
	}
}

/*
 * Suspend tip
 */
suspend(c)
	char c;
{

	tio_unraw();
	kill(c == CTRL('y') ? getpid() : 0, SIGTSTP);
	tio_raw();
}

/*
 *	expand a file name if it includes shell meta characters
 *
 * 1995/12
 * Kenar says:
 * It is doubtful to support shell meta-char so rigorously.
 * So I replaced original 'expand' by very simple one
 * that support only '~'(HOME).
 * -Kenar-
 */

char *expand(name) char *name;
{	char *s;
	static char buf[PATHSIZ];
	s = buf;
	*s = '\0';
	if(*name == '~'){ /* then we replce it HOME */
		s = strapp(s,svalue(HOME));
		name++;
	}
	s = strapp(s, name);
	return buf;
}


/*
 * Are any of the characters in the two strings the same?
 */

anyof(s1, s2)
	register char *s1, *s2;
{
	register int c;

	while (c = *s1++)
		if (any(c, s2))
			return(1);
	return(0);
}
