/*
 * Copyright (c) 1983 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Berkeley.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/*
 * Copyright (c) 1995  Kenji Arisawa. All right reserved.
 * License to freely use and distribute this software is hereby granted 
 * by the author, subject to the condition that this copyright notice 
 * remains intact.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY WARRANTIES.
 */
 

#ifndef lint
static char sccsid[] = "@(#)hunt.c	5.4 (Berkeley) 9/2/88";
#endif /* not lint */

#include "tip.h"

boolean lockflag = NO;

static int fd;
static	jmp_buf deadline;

sigfunc_t dead()
{
	fd = -1;
	longjmp(deadline, 1);
}

/*
 * hunt: hunt device to be opened. 
 * 1995/12 tip+ mod. by Kenar
 * 1996/01 Rewritten by Kenar
 *
 * Tip+ new usage is,
 * return:
 *    file descripter that is opend.
 *    ERR: not found or opened.
 * Hunt now cleans locking file if failed.
 * EXAMPLE
 * Let host is given in remote as bellow.
 *    host:\
 *       dv=/dev/ttya,/dev/ttyb,/dev/ttyc
 * and let /dev/ttya is being used.
 * Then hunt("host") will return the value open("/dev/ttyb",O_RDWR).
 * -Kenar-
 */
int hunt(name,baudopt)
	char *name;
	int baudopt; /* command baudrate option */
{
	char *cp;
	sigfunc_t (*oldf)();
	int bcode;

	oldf = signal(SIGALRM, dead);
	getremcap(name);
	
   /*
	* examine the validity of baudrate prior to open the device
	* -Kenar-
	*/
	if(baudopt) nvalue(BAUDRATE) = baudopt;
	if((bcode = baudcode(nvalue(BAUDRATE))) == 0){
		printf("tip: bad baud rate %d. Select from:\n",nvalue(BAUDRATE));
		showbauds();
		exit(3);
	}
	
   /*
    * All devices in DV are examined if it can be open.
	* DV example: DV="/dev/ttya,/dev/ttyb,/dev/ttyc".
	* In return open device is set to DV, for example: DV="/dev/ttyb"
	* -Kenar-
	*/
	for(cp = getoken(DV,","); cp; cp = getoken(NULL,",")){
 		if (uu_lock(cp) < 0)
			continue;
		/*
		 * Straight through call units, such as the BIZCOMP,
		 * VADIC and the DF, must indicate they're hardwired in
		 *  order to get an open file descriptor placed in FD.
		 * Otherwise, as for a DN-11, the open will have to
		 *  be done in the "open" routine.
		 */
		 
		if (setjmp(deadline) == 0) {
			alarm(10);
			fd = open(cp, O_RDWR);
			/* What is the value fd, if timeout ?
			 * It is obscure in the original tip code. ( Buggy )
			 * I set it -1 in dead().
			 * -Kenar-
			 */
		}
		alarm(0);
		
		if (fd < 0) {
			perror(cp);
		}
		else {
		   /* --- tip+ ---
			* We have cases fd > 0 and not really connecting
			* 1. Hardware wired.
			* 2. Modem is power off.
			* 3. NxFaxMonitor(Black&White co.) is running.
			* -Kenar-
			*/
			lockflag = YES;
			setxlock(fd); /* Set exclusive use of tty */
			sethup(fd); /* Hang up on last close. (disconnect modem line) */
			setdtr(fd); /* Set data terminal ready */
						/* May be unnecessary -Kenar- */
			
		   /*
			*  Hardwired connections require the
			*  line speed set before they make any transmissions
			*  (this is particularly true of things like a DF03-AC)
			*/
			ttysetup(fd,bcode,bvalue(TAND));
			signal(SIGALRM, oldf);
			printf("connected to %s\n",cp);
			DV = makstr(cp);
			return (fd);
		}
		uu_unlock();
	}
	signal(SIGALRM, oldf);
	return ERR;
}
