/*
 * Copyright (c) 1995  Kenji Arisawa. All right reserved.
 * License to freely use and distribute this software is hereby granted 
 * by the author, subject to the condition that this copyright notice 
 * remains intact.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY WARRANTIES.
 */
 #ifndef lint
static char sccsid[] = "@(#)ioctl.c	1.0 (Kenar) 18/5/96";
#endif /* not lint */

#include <sys/ioctl.h>
#ifdef USE_IOCTL_COMPT
#include <sys/ioctl_compat.h>
#endif

#define YES 1
#define NO 0

typedef int boolean;


boolean rawflag = NO; /* Is in rawmode ? */


/* terminal data */
static struct sgttyb	arg;		/* current mode of local terminal */
static struct sgttyb	defarg;		/* initial mode of local terminal */
static struct tchars	tchars;		/* current state of terminal */
static struct tchars	defchars;	/* initial state of terminal */

static struct ltchars	ltchars;	/* current local characters of terminal */
static struct ltchars	deflchars;	/* initial local characters of terminal */
static int	odisc;				/* initial tty line discipline */
static int disc = OTTYDISC;		/* tip normally runs this way */

/* line data */
static struct sgttyb ttyarg;


/* ------------ tty control functions -------------------*/

/*
 * put the controlling keyboard into raw mode
 */
void tio_raw()
{	static boolean done = NO;

	if(!done){
		ioctl(0,TIOCGETP,(char *)&defarg);
		ioctl(0,TIOCGETC,(char *)&defchars);
		ioctl(0,TIOCGLTC,(char *)&deflchars);
		ioctl(0,TIOCGETD,(char *)&odisc);
		arg = defarg;
		
		/* tip+ */
		arg.sg_flags = CBREAK;
		/* What we have if we set
		 *	arg.sg_flags = RAW|CBREAK
		 * ? -Kenar- 
		 */
		/*
		tip+: ---    arg.sg_flags = CBREAK;
		tipx: ---    arg.sg_flags = (rawthru?RAW:CBREAK);
		tip:  ---	 arg.sg_flags = ANYP | CBREAK;
		* If we use RAW for sg_flags then we must make changes to
		* tio_cbreak()
		* to enable CBREAK.
		* -Kenar-
		*/
	
		tchars = defchars;
		tchars.t_intrc = tchars.t_quitc = -1;
		ltchars = deflchars;
		ltchars.t_suspc = ltchars.t_dsuspc = ltchars.t_flushc
			= ltchars.t_lnextc = -1;
		done = YES;
	}
	ioctl(0,TIOCSETP,&arg);
	ioctl(0,TIOCSETC,&tchars);
	ioctl(0,TIOCSLTC,&ltchars);
	ioctl(0,TIOCSETD,(char *)&disc);
	rawflag = YES;
}


/*
 * return keyboard to normal mode
 */
void tio_unraw()
{
	ioctl(0,TIOCSETD,(char *)&odisc);
	ioctl(0,TIOCSETP,(char *)&defarg);
	ioctl(0,TIOCSETC,(char *)&defchars);
	ioctl(0,TIOCSLTC,(char *)&deflchars);
	rawflag = NO;
}

/* set/clear cbreak */
void tio_cbreak(mode)
boolean mode;
{	if(mode) ioctl(0, TIOCSETC, &defchars);
	else ioctl(0, TIOCSETC, &tchars);
}


/* set/clear tandem */
void tio_tandem(mode)
boolean mode;
{		if(mode) arg.sg_flags |= TANDEM;
		else arg.sg_flags &= ~TANDEM;
		ioctl(0, TIOCSETP, &arg);
}


/* ---------------- ioctrl for modem -------------------- */

/*
 * Set up the "remote" tty's state
 */
void ttysetup(fd,baudcode,tandmode)
int fd;
int baudcode; /* speed is given by baud code - Kenar - */
boolean tandmode; /* tandem mode */
{   /* ref: <sys/ioctl.h> for NeXT */
	unsigned bits = LDECCTQ; /* only ^Q starts after ^S */

	ttyarg.sg_ispeed = ttyarg.sg_ospeed = baudcode;
	ttyarg.sg_flags = RAW; /* no i/o processing */
	if(tandmode) ttyarg.sg_flags |= TANDEM; /* send stopc on out q full */
	else ttyarg.sg_flags &= ~TANDEM;
	ioctl(fd,TIOCSETP,(char *)&ttyarg);
	ioctl(fd,TIOCLBIS,(char *)&bits);
}

void clrdtr(fd) /* clear data terminal ready */
int fd;
{
	ioctl(fd, TIOCCDTR, 0);
}

void setdtr(fd) /* set data terminal ready */
int fd;
{
	ioctl(fd, TIOCSDTR, 0);
}

void iflush(fd) /* flush input buf to remove echo garbage */
int fd;
{
	ioctl(fd, TIOCFLUSH, 0);
}

int nread(fd) /* get data size to read */
int fd;
{	int n;
	ioctl(fd, FIONREAD, &n);
	return n;
}

void setxlock(fd)  /* Set exclusive use of tty */
int fd;
{
	ioctl(fd, TIOCEXCL, 0);
}

void sethup(fd) /* Hang up on last close. (disconnect modem line) */
int fd;
{
	ioctl(fd, TIOCHPCL, 0);
}

void sendbrk(fd,msec) /* send break ( mili second ) */
int fd;
int msec;
{
	ioctl(fd, TIOCSBRK, 0);
	nap(msec);
	ioctl(fd, TIOCCBRK, 0);
}

/* set/clear tandem */
void io_tandem(fd,cntl)
int fd;
boolean cntl;
{	struct sgttyb rmtty;
	ioctl(fd, TIOCGETP, &rmtty);
	if(cntl) rmtty.sg_flags |= TANDEM;
	else rmtty.sg_flags &= ~TANDEM;
	ioctl(fd, TIOCSETP, &rmtty);
}

/* ---------------- file lock  -------------------- */
