/*
 * Copyright (c) 1993  Kenji Arisawa. All right reserved.
 * License to freely use and distribute this software is hereby granted 
 * by the author, subject to the condition that this copyright notice 
 * remains intact.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY WARRANTIES.
 */

/* miscellaneous functions added by Kenar */
#ifndef lint
static char sccsid[] = "@(#)misc.c	1.0 (Kenar) 2/3/96";
#endif /* not lint */

#include <time.h>
#include "tip.h"
#define ERR (-1)


char *strapp(t,s) char *t,*s;
{	while(*s) *t++ = *s++;
	*t = '\0';
	return t;
}


/*
 * skip: skip chars.
 * EXAMPLES:
 *   skip("  abc"," ") returns "abc".
 *   skip("   "," ") returns "".
 * Never return NULL;
 * -Kenar-
 */
char *skip(s,str)
char *s,*str;
{	while(*s && index(str,*s)) s++;
	return s;
}


/*
 * skipto: skip to chars.
 * EXAMPLES:
 *   skipto("abc  de"," ") returns " de".
 *   skipto("abc"," ") returns "".
 *   skipto("  "," ") returns "  ".
 *   skipto(""," ") returns "".
 * Never return NULL;
 * -Kenar-
 */
char *skipto(s,str)
char *s,*str;
{	while(*s && !index(str,*s)) s++;
	return s;
}


/*
 * arg0: return argv[0] if s is command line
 * In case blank line such as s = "    ",
 * arg0(s) returns "" ( that is, never return NULL )
 * arg0 make no side effect.
 * -Kenar-
 */
char *arg0(s)
char *s;
{	static char buf[STRSIZ];
	char *sp = " \t\n";
	char *t;
	strcpy(buf,s);
	s = skip(buf,sp);
	t = skipto(s,sp);
	*t = '\0';
	return s;
}

char *trim(buf)
char *buf;
{	char *tp,*p;
	char *spaces=" \t\n";
	tp = skip(buf,spaces);
	p = tp + strlen(tp);
	while(index(spaces,*p)) p--;
	p++;
	*p = '\0';
	strcpy(buf,tp);
	return buf;
}


/*
 * oneof: test name matching.
 * return:
 *   1,2,3... if match
 *   0        if fail
 * EXAMPLE
 *    if(oneof(person,"ken,fumi,yu,sho")) puts("My family")
 * -Kenar-
 */
int oneof(s,r)
	char *s;
	char *r;
{	int n = 0;
	int ok = 0;
	char *t;
	char buf[256];
	strcpy(buf,r);
	r = buf;
	do{
		t = skipto(r,",");
		*t++ = '\0';
			if(strcmp(s,r) == 0){
			ok = 1;
			break;
		}
		n++;
		r = t;
	}while(*r);
	if(!ok) n = -1;
	return (n + 1);
}

/*
 * getoken: get token
 * getting token is a troublesome process.
 * The last of string brings sometimes bugs.
 * So I made a general function...
 * -Kenar-
 */
char *getoken(bp,sep)
char *bp,*sep;
/*
 * Let bp="abc  def" and sep=" ", then
 * 1'st try returns "abc"
 * 2'nd try returns "def"
 * 3'rd try returns NULL
 * Only the first try we put bp else put NULL.
 */
{	static char buf[STRSIZ];
	static char *s,*t;
	
	if(bp){
		s = strcpy(buf,bp);
		buf[strlen(buf) + 1] = '\0'; /* important */
	}
	else
		s = t;
	s = skip(s,sep);
	if(!*s) return NULL;
	t = skipto(s,sep);
	*t++ = '\0';
	return s;
}

	
/*
int isblank(s) char *s;
{	char c;
	while(*s)
	{	c = *s & 0x8f;
		if(isspace(c)) s++;
		else return 0;
	}
	return 1;
}
*/

char *strlower(s) char *s;
{	char c,*p;
	p = s;
	while(*p)
	{	c = *p & 0x7f;
		*p = (isupper(c)?tolower(c):*p);
		p++;
	}
	return s;
}

char *strupper(s) char *s;
{	char c,*p;
	p = s;
	while(*p)
	{	c = *p & 0x7f;
		*p = (islower(c)?toupper(c):*p);
		p++;
	}
	return s;
}


int sepstr(str,argp,sep)
char *str,**argp,*sep;
/*
   separate string to arguments
	... coded by Kenar
   sep: separaters
   return: argument count
*/
{	int n;
	n = 0;
	for(;;)
	{	while(index(sep,*str)) str++;
		if(!*str) break;
		argp[n++] = str;
		while(!index(sep,*str) && *str) str++;
		if(!*str) break;
		*str++ = 0;
	}
	argp[n] = NULL;
	return n;
}

/*
*	inmem: test if char is in memory
*/
boolean inmem(c,bp,siz)
char c;
char *bp;
int siz;
{	while(siz && (c != *bp++)) siz--;
	if(siz) return YES;
	else return NO;
}

/*
*	chcount: count char in memory
*/
int chcount(c,bp,siz)
char c;
char *bp;
int siz;
{	int count;
	count = 0;
	while(siz){
		if(c == *bp++) count++;
		siz--;
	}
	return count;
}



char *jdate()
/*	jdate returns local time in the form:
 *	YYMMDD
 */
{	time_t secs;
	struct tm *t;
	static char buf[7];
	time(&secs);
	t = localtime(&secs);
	/*puts(asctime(t));*/
	sprintf(buf,"%02d%02d%02d",
		t->tm_year, t->tm_mon+1, t->tm_mday);
	return buf;
}
