/*
 * Copyright (c) 1988 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Berkeley.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef lint
static char sccsid[] = "@(#)uucplock.c	5.3 (Berkeley) 9/2/88";
#endif /* not lint */

#include "tip.h"

/* Moved to config.h -Kenar-
#define LOCKDIRFMT	"/usr/spool/uucp/LCK/LCK..%s"
*/

static char *lockfile = NULL;


/*
 * The name of the lock is pretty important. We don't
 * want to introduce the following security problems:
 * 1. Namecollision in lock file between different devices
 * 2. uucplock is used to decide whether or not to call
 *   acucntrl. Thus, non-/dev devices should have a different
 *    lock name than devices in /dev.
 */
/*
 * At this point we are guranteed that any devices outside
 * /dev will have a lock name starting with underscore (_),
 * and that any slashes (/) in the pathname have been changed
 * to underscores, and that multiple slashes are only made
 * into a single underscore. Thus, any way of expressing
 * a pathname will lead to a unique lock file name.
 */
/*
 * 1996/01 Moved from hunt.c and defined new function. -Kenar-
 */
char *mklockname(cp) char *cp;
{	char *cf,*ct;
	char buf[PATHSIZ];
	
	if (*cp != '/')	/* device name must be absolute path */
		return NULL;
	while (*cp == '/') cp++;
	cp--;
	if (strncmp("/dev/", cp, 5) == 0) {
		cf = cp + 5;
		while (*cf == '/') cf++;
	} else
		cf = cp;
	ct = buf;
	while (*cf) {
		if (*cf == '/') {
			if (ct == buf
				|| (ct > buf && *(ct-1) != '_'))
				*ct++ = '_';
		} else
			*ct++ = *cf;
		cf++;
	}
	*ct = '\0';
	return makstr(buf);
}


/* 
 * uucp style locking routines
 * return:
 *     0 - success
 * 	  -1 - failure
 */
/*
 * 1996/01 Kenar: A little correction.
 */
int uu_lock(ttyname)
	char *ttyname;
{
	extern int errno;
	int fd, pid;
	char tbuf[PATHSIZ];
	off_t lseek();

	 /* 1996/01: inserted */
	lockfile = mklockname(ttyname);
	if(!lockfile) return ERR;
	
	(void)sprintf(tbuf, LOCKDIRFMT, lockfile);
	fd = open(tbuf, O_RDWR|O_CREAT|O_EXCL, 0660);
	if (fd < 0) {
		/*
		 * file is already locked
		 * check to see if the process holding the lock still exists
		 */
		fd = open(tbuf, O_RDWR, 0);
		if (fd < 0) {
			perror("lock open");
			return(ERR);
		}
		if (read(fd, &pid, sizeof(pid)) != sizeof(pid)) {
			(void)close(fd);
			perror("lock read");
			return(ERR);
		}

		if (kill(pid, 0) == 0 || errno != ESRCH) {
			(void)close(fd);	/* process is still running */
			return(ERR);
		}
		/*
		 * The process that locked the file isn't running, so
		 * we'll lock it ourselves
		 */
		if (lseek(fd, 0L, L_SET) < 0) {
			(void)close(fd);
			perror("lock lseek");
			return(ERR);
		}
		/* fall out and finish the locking process */
	}
	pid = getpid();
	if (write(fd, (char *)&pid, sizeof(pid)) != sizeof(pid)) {
		(void)close(fd);
		(void)unlink(tbuf);
		perror("lock write");
		return(ERR);
	}
	(void)close(fd);
	return(0);
}

/*
 * 1996/01 Kenar: no return value.
 */
void uu_unlock()
{
	char tbuf[PATHSIZ];
	if(!lockfile) return;
	(void)sprintf(tbuf, LOCKDIRFMT, lockfile);
	if(unlink(tbuf)) fprintf(stderr,"#Error when unlink %s",tbuf);
}
