/*
 * Copyright (c) 1983 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Berkeley.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef lint
static char sccsid[] = "@(#)value.c	5.2 (Berkeley) 9/2/88";
#endif /* not lint */

#include "tip.h"

#define MIDDLE	35

extern int pid;

static char *vtoken(char *s);
static void vprint(int n);
static int vlookup(char *s);
static char *intstr(char *buf,char *s,char *sep);
static char *interp(char *s);
static int col = 0;


/*
 * vassign: assign value to variable and put CHANGE flag
 * original usage is changed by Kenar.
 * new usage is,
 * EXAMPLES
 * vassign(BAUDRATE,"9600")
 * vassign(RAISECHAR,"\001")
 *
 * 1995/12 tip+ mod. by Kenar
 * 1996/02 tip+ mod. by Kenar
 * 1996/07 tip+ mod. by Kenar
 * -Kenar-
 */
static void vassign(id, v)
	int id;
	char *v;
{	char buf[STRSIZ];

	if (!(vaccess(id)&WRITE)) {
		printf("%s access denied\r\n",vname(id));
		return;
	}
	switch (vtype(id)&TMASK) {

	case STRING:
		if (equal(svalue(id), v)) return;
		svalue(id) = makstr(v);
		break;

	case NUMBER:
		if (nvalue(id) == atoi(v)) return;
		nvalue(id) = atoi(v);
		break;

	case BOOL:
		strcpy(buf,v); /* To avoid bass error */
		strupper(buf);
		if (equal(buf,"ON")||equal(buf,"OFF")){
			if (bvalue(id) == equal(buf,"ON")) return;
			bvalue(id) = equal(buf,"ON");
		}
		else{
			printf("%s: Use ON/OFF for boolean value\r\n",v);
			return;
		}
		break;
	case CHAR:
		if (cvalue(id) == *v) return;
		cvalue(id) = *v;
		break;
	}
	vaccess(id) |= CHANGED;
}



/*
 * vcheck: remove CHANGE mark, do some actions for change,
 * and send message to child if he exist.
 * vcheck follows vlex.
 * Note that "set record=blah script" do work correctly,
 * because RECORD < SCRIPT.
 *
 * vcheck should not be executed in child process.
 *
 * -Kenar-
 */
void vcheck()
{	int n;
	int err = NO;
	char *s,*t;

	for(n = 0; vname(n); n++){
		if(vaccess(n)&CHANGED){
			vaccess(n) &= ~CHANGED;
			/* and we have some exceptions */
			switch(n){
			case TAND: tandem(bvalue(TAND)); break;
			case PARITY:
				err = setpio(svalue(PARITY));
				break;
			case LANGTYPE:
				err = setpio(svalue(LANGTYPE));
				break;
			case LDELAY:
				if(nvalue(LDELAY) <0) nvalue(LDELAY) = 0;
				if(nvalue(LDELAY) > 999) nvalue(LDELAY) = 999;
				break;
			case CDELAY:
				if(nvalue(CDELAY) <0) nvalue(CDELAY) = 0;
				if(nvalue(CDELAY) > 999) nvalue(CDELAY) = 999;
				break;
			case KEYMAP:
				if(vdef(KEYMAP)){
					s = svalue(KEYMAP);
					if(strlen(s) != 3 || s[1] != ':'){
						puts("Format is X:Y where X,Y are characters.\r");
						vundef(KEYMAP);
						err = ERR;
					}
				}
				break;
			case RECORD:
				s = expand(svalue(RECORD));
				if((t = index(s,'*')) != NULL){
					/* then replace '*' by YYMMDD
					 * Don't use '?' instead of '*', because '?' is a token
					 * in ~s command.
					 * 1996/07: Kenar
					 */
					char buf[STRSIZ];
					*t++ = 0;
					sprintf(buf,"%s%s%s",s,jdate(),t);
					s = makstr(buf); /* don't move out of scope */
				}
				else s = makstr(s);
				svalue(RECORD) = s;
				break;
			default:;
			}
			if(!err && pid){ /* Let child know the change. See tipout.c */
				sigEMT(n);
				nap(100); /* sleep 100ms */
			}
		}
	}
}


/*
 * vlex: set or print variables complying with a input of [set] command.
 * vlex is used in precedence to vcheck.
 * variable definition in tip+ script (system and/or users), in '.tiprc'
 * and variable setting in [set] command uses vlex.
 * EXAMPLES
 * vlex("all") prints all tip variables
 * vlex("?") prints variables of WRITE permission.
 * vlex(" W?  X=a   Y=b !Z ")
 * 
 * 1995/12 Kenar: slight modification for readability.
 * 1996/01 Kenar: I make things simple.
 *   1. Bug fix in multi-definitions.
 *   2. any spaces support.
 * -Kenar-
 */
void vlex(s) char *s;
{
	int n;
		
	s = trim(s);
	if(equal(s, "all")) {
		for (n = 0; vname(n); n++)
			if (vaccess(n)&READ) vprint(n);
	} else if(equal(s,"?")) {
		for (n = 0; vname(n); n++)
			if (vaccess(n)&WRITE) vprint(n);
	} else {
		s[strlen(s) + 1] = '\0'; /* A magic to make program easy */
		repeat
			s = vtoken(s);
		until(!s);
	}
	if (col > 0) {
		printf("\r\n");
		col = 0;
	}
}

char *skipword(s) char *s;
{	if(*s && isalpha(*s)) s++;
	while(*s && isalnum(*s)) s++;
	return s;
}

/*
 * vtoken: set or print value of single variable.
 * EXAMPLE
 * vtoken("TOKEN") where we allow for TOKEN:
 * NAME=abc
 * NAME="cat %s"
 * NAME          --- old style for BOOL
 * !NAME         --- old style for BOOL
 * NAME=ON       --- new style for BOOL
 * NAME=OFF      --- new style for BOOL
 * NAME?
 *
 * RETURN
 *   NULL: if we should terminate
 *   else: if we should look up more. Next pointer to look up.
 * 1996/02	recontruction by Kenar
 * -Kenar-
 */
static char *vtoken(s)
	char *s;
{
	int n;
	char *t;
	char buf[STRSIZ];
	char *sep=" \t";

	s = skip(s,sep);
	if(!*s) return NULL;
	if(*s == '!'){
		s++;
		t = skipword(s);
		if(*t && !index(sep,*t)){
			printf("%s: Error in format\r\n",s);
			return NULL;
		}
		*t = '\0';
		n = vlookup(s);
		if (n == ERR){
			printf("%s: No such variable\r\n",s);
			return NULL;
		}
		if(vtype(n)&BOOL) vassign(n, "OFF");
		else{
			printf("%s: Error in type\r\n",s);
			return NULL;
		}
		return ++t;
	}
	
	t = skipword(s);
	if(!*t || index(sep,*t)){ /* Then case BOOL */
		*t = '\0';
		n = vlookup(s);
		if (n == ERR){
			printf("%s: No such variable\r\n",s);
			return NULL;
		}
		if(vtype(n)&BOOL) vassign(n, "ON");
		else{
			printf("%s: Error in type\r\n",s);
			return NULL;
		}
		return ++t;
	}

	if(*t == '?') { /* Then case NAME? (Print variable) */
		*t = '\0';
		if ((n = vlookup(s)) != ERR && (vaccess(n)&READ)) {
			vprint(n);
			return ++t;
		}
		else {
			printf("%s: No such variable\r\n",s);
			return NULL;
		}
	}
	
	/* Then case NAME=String or NAME="string" */
	if(*t != '='){
		printf("%s: Error in format\r\n",s);
		return NULL;
	}
	*t = '\0';
	n = vlookup(s);
	if (n == ERR){
		printf("%s: No such variable\r\n",s);
		return NULL;
	}
	t++;
	if(*t == '\"'){ /* Then we are in "..." */
		t++;
		s = t;
		t = intstr(buf,s,"\"");
		if(*t != '\"'){
			printf("%s: Error in format\r\n",s);
			return NULL;
		}
		vassign(n, buf);
		return ++t;
	}
	s = t;
	t = intstr(buf,s,sep);
	if(*t && !index(sep,*t)){
		printf("%s: Error in format\r\n",s);
		return NULL;
	}
	vassign(n, buf);
	return t;
}


/*
 * print value of tip variable.
 *
 * 1995/12	slight modification for readablity by Kenar
 * -Kenar-
 */
static void vprint(n) int n;
{
	char *cp;

	if (col > 0 && col < MIDDLE)
		while (col++ < MIDDLE)
			putchar(' ');
	col += size(vname(n));
	switch (vtype(n)&TMASK) {

	case BOOL:
		col += 3;
		if (bvalue(n) == NO) {
			printf("%s=OFF", vname(n));
			col++;
		}
		else printf("%s=ON", vname(n));
		break;

	case STRING:
		printf("%s=", vname(n));
		col++;
		cp = interp(svalue(n));
		col += size(cp);
		if(index(cp,' ')){
			printf("\"%s\"", cp);
			col += 2;
		}
		else printf("%s", cp);
		break;

	case NUMBER:
		col += 6;
		printf("%s=%-5d", vname(n), nvalue(n));
		break;

	case CHAR:
		printf("%s=", vname(n));
		col++;
		cp = ctrl(cvalue(n));
		col += size(cp);
		printf("%s", cp);
		break;
	}
	if (col >= MIDDLE) {
		col = 0;
		printf("\r\n");
		return;
	}
}


/*
 * look up tip variable using name.
 * return: variable id.
 * EXAMPLE
 * vlookup("ba") returns BAUDRATE
 *
 * 1995/12	slight modification for readablity by Kenar
 * -Kenar-
 */
static int vlookup(s) char *s;
{
	int n;

	for (n = 0; vname(n); n++)
		if (equal(vname(n), s) || (vabbre(n) && equal(vabbre(n), s)))
			return (n);
	return (ERR);
}

/*
 * vinterp: convert string such like "\t\e\b\n" to it's internal code.
 *
 * 1996/01 Kenar
 * 2. \e support as interp() do.
 * 3. chaged: vinterp(s,stop)  ->  vinterp(s)
 * -Kenar-
 */
static char *intstr(buf,s, sep)
	char *buf,*s,*sep;
{	char *p, c;
	int num;
	p = buf;
	while (*s && !index(sep,*s)){
		c = *s++;
		switch (c) {
		case '^':
			if (*s == '?'){
				*p++ = 0177; /* DEL code (tip+) */
				s++;
			}
			else if (*s)
				*p++ = *s++ - 0100;
			else
				*p++ = c;
			break;

		case '\\':
			num = 0;
			c = *s++;
			if (c >= '0' && c <= '7')
				num = (num<<3)+(c-'0');
			else {
				register char *q = "n\nr\rt\tb\bf\fe\033E\033\"\"";

				for (; *q; q++)
					if (c == *q++) {
						*p++ = *q;
						goto cont;
					}
				*p++ = c;
			cont:
				break;
			}
			if ((c = *s++) >= '0' && c <= '7') {
				num = (num<<3)+(c-'0');
				if ((c = *s++) >= '0' && c <= '7')
					num = (num<<3)+(c-'0');
				else
					s--;
			} else
				s--;
			*p++ = num;
			break;

		default:
			*p++ = c;
		}
	}
	*p = '\0';
	return s;
}

/*
 * interp: convert string to printable format. The dual to vinterp().
 *
 * 1996/01 Kenar
 * 1. Moved from tip.c
 * 2. changed expression from \E to \e
 * -Kenar-
 */
static char *interp(s)
	char *s;
{
	static char buf[256];
	register char *p = buf,c,*q;

	while(c = *s++)
	{
		for(q = "\nn\rr\tt\ff\033e\bb\"\""; *q; q++)
			if(*q++ == c)
			{
				*p++ = '\\';
				*p++ = *q;
				goto next;
			}
		if(c < 040)
		{
			*p++ = '^';
			*p++ = c + 'A'-1;
		}
		else if(c == 0177)
		{
			*p++ = '^';
			*p++ = '?';
		}
		else
			*p++ = c;
next:
		;
	}
	*p = '\0';
	return(buf);
}
